/* 
 * This file is part of PLaSK (https://plask.app) by Photonics Group at TUL
 * Copyright (c) 2022 Lodz University of Technology
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 */
#ifndef PLASK__UTILS_EVENT_H
#define PLASK__UTILS_EVENT_H

/** @file
This file contains utils for events types.
*/

namespace plask {

/**
 * Base class for classes which store information about event generated by objects of type @p SourceType.
 *
 * It contains flags field which describes additional event properties.
 * Subclasses of subclasses of this can also contains additional information about specific type of event.
 */
template <typename SourceType, typename _FlagsType = unsigned char>
class EventWithSourceAndFlags {

    /// Source of event.
    SourceType* _source;

    /// Event flags (which describes event properties).
    _FlagsType _flags;

public:

    typedef _FlagsType FlagsType;

    /**
     * Get source of event.
     * @return source of event
     */
    const SourceType* source() const { return _source; }

    /**
     * Get source of event.
     * @return source of event
     */
    SourceType* source() { return _source; }

    /**
     * Get source of event casted (using dynamic_cast) to given type T.
     * @return source of event as T
     * @tparam T required type
     */
    template <typename T>
    const T* source() const { 
        const T* casted = dynamic_cast<const T*>(_source);
        if (!casted && _source) throw std::bad_cast();
        return casted;
    }

    /**
     * Get source of event casted (using dynamic_cast) to given type T.
     * @return source of event as T
     * @tparam T required type
     */
    template <typename T>
    T* source() {
        T* casted = dynamic_cast<T*>(_source);
        if (!casted && _source) throw std::bad_cast();
        return casted;
    }

    /**
     * Get event's flags.
     * @return flags which describes event's properties
     */
    FlagsType flags() const { return _flags; }

    /**
     * Get event's flags with excluded @p flagsToRemove.
     * @param flagsToRemove flags to exclude from result
     * @return flags which describes event properties without @p flagsToRemove
     */
    FlagsType flagsWithout(FlagsType flagsToRemove) const { return _flags & ~flagsToRemove; }

    /**
     * Check if given all of @p flags are set.
     * @param flags flags to check
     * @return @c true only if all of @p flags are set
     */
    bool hasAllFlags(FlagsType flags) const { return (_flags & flags) == flags; }

    /**
     * Check if one of given @p flags are set.
     * @param flags flags to check
     * @return @c true only if one or more of @p flags are set
     */
    bool hasAnyFlag(FlagsType flags) const { return (_flags & flags) != 0; }

    /**
     * Construct event.
     * @param source source of event
     * @param flags falgs which describes event's properties
     */
    explicit EventWithSourceAndFlags(SourceType* source, FlagsType flags = 0): _source(source), _flags(flags) {}

    /// Virtual destructor for eventual subclassing
    virtual ~EventWithSourceAndFlags() {}

};

}   // namespace plask

#endif // PLASK__UTILS_EVENT_H
